@@ -1,10 +1,17 @@
 #include "system.h"
 
 #include <stdarg.h>
-#if defined(__linux__) && defined(__powerpc__)
-#include <setjmp.h>
+
+#if defined(__linux__)
+#include <elf.h>
+#include <link.h>
 #endif
 
+
+#if HAVE_SYS_UTSNAME_H
+#include <sys/utsname.h>
+#endif
+#include <netdb.h>
 #include <ctype.h>	/* XXX for /etc/rpm/platform contents */
 
 #if HAVE_SYS_SYSTEMCFG_H
@@ -68,12 +75,15 @@ struct rpmOption {
     char * name;
     int var;
     int archSpecific;
-int required;
     int macroize;
     int localize;
-struct rpmOptionValue * value;
 };
 
+static struct rpmat_s {
+    const char *platform;
+    uint64_t hwcap;
+} rpmat;
+
 typedef struct defaultEntry_s {
     char * name;
     char * defName;
@@ -117,12 +127,12 @@ static struct tableType_s tables[RPM_MACHTABLE_COUNT] = {
 #define RPMVAR_NUM                      55      /* number of RPMVAR entries */
 
 /* this *must* be kept in alphabetical order */
-/* The order of the flags is archSpecific, required, macroize, localize */
+/* The order of the flags is archSpecific, macroize, localize */
 
-static const struct rpmOption const optionTable[] = {
-    { "include",		RPMVAR_INCLUDE,			0, 1,	0, 2 },
-    { "macrofiles",		RPMVAR_MACROFILES,		0, 0,	0, 1 },
-    { "optflags",		RPMVAR_OPTFLAGS,		1, 0,	1, 0 },
+static const struct rpmOption optionTable[] = {
+    { "include",		RPMVAR_INCLUDE,			0, 0, 2 },
+    { "macrofiles",		RPMVAR_MACROFILES,		0, 0, 1 },
+    { "optflags",		RPMVAR_OPTFLAGS,		1, 1, 0 },
 };
 
 static const size_t optionTableSize = sizeof(optionTable) / sizeof(*optionTable);
@@ -160,7 +170,7 @@ machCacheFindEntry(const machCache cache, const char * key)
     int i;
 
     for (i = 0; i < cache->size; i++)
-	if (!strcmp(cache->cache[i].name, key)) return cache->cache + i;
+	if (rstreq(cache->cache[i].name, key)) return cache->cache + i;
 
     return NULL;
 }
@@ -337,7 +347,6 @@ static rpmRC addCanon(canonEntry * table, int * tableLen, char * line,
 	return RPMRC_FAIL;
     }
 
-   	/* LCL: s != NULL here. */
     tnum = strtoul(s, &s1, 10);
     if ((*s1) || (s1 == s) || (tnum == ULONG_MAX)) {
 	rpmlog(RPMLOG_ERR, _("Bad arch/os number: %s (%s:%d)\n"), s,
@@ -392,7 +401,7 @@ static canonEntry lookupInCanonTable(const char * name,
 {
     while (tableLen) {
 	tableLen--;
-	if (strcmp(name, table[tableLen].name))
+	if (!rstreq(name, table[tableLen].name))
 	    continue;
 	return &(table[tableLen]);
     }
@@ -406,7 +415,7 @@ const char * lookupInDefaultTable(const char * name,
 {
     while (tableLen) {
 	tableLen--;
-	if (table[tableLen].name && !strcmp(name, table[tableLen].name))
+	if (table[tableLen].name && rstreq(name, table[tableLen].name))
 	    return table[tableLen].defName;
     }
 
@@ -427,6 +436,7 @@ static void setDefaults(void)
     if (!macrofiles) {
 	macrofiles = rstrscat(NULL, confdir, "/macros", ":",
 				confdir, "/platform/%{_target}/macros", ":",
+				confdir, "/fileattrs/*.attr", ":",
   				confdir, "/" RPMCANONVENDOR "/macros", ":",
 				SYSCONFDIR "/rpm/macros.*", ":",
 				SYSCONFDIR "/rpm/macros", ":",
@@ -501,26 +511,18 @@ static rpmRC doReadRC(const char * urlfn)
 		goto exit;
 	    }
 
-	    switch (option->var) {
-	    case RPMVAR_INCLUDE:
+	    if (option->var == RPMVAR_INCLUDE) {
 		s = se;
 		while (*se && !risspace(*se)) se++;
-		if (*se != '\0') *se++ = '\0';
-
-#if 0 /* XXX doesn't seem to do anything useful, only break things... */
-		rpmRebuildTargetVars(NULL, NULL);
-#endif
+		if (*se != '\0') *se = '\0';
 
 		if (doReadRC(s)) {
 		    rpmlog(RPMLOG_ERR, _("cannot open %s at %s:%d: %m\n"),
 			s, fn, linenum);
 		    goto exit;
-		} else {
-		    continue;	/* XXX don't save include value as var/macro */
 		}
-	      	break;
-	    default:
-		break;
+		/* XXX don't save include value as var/macro */
+		continue;
 	    }
 
 	    if (option->archSpecific) {
@@ -546,7 +548,7 @@ static rpmRC doReadRC(const char * urlfn)
 
 	    /* Only add macros if appropriate for this arch */
 	    if (option->macroize &&
-	      (arch == NULL || !strcmp(arch, current[ARCH]))) {
+	      (arch == NULL || rstreq(arch, current[ARCH]))) {
 		char *n, *name;
 		n = name = xmalloc(strlen(option->name)+2);
 		if (option->localize)
@@ -565,7 +567,7 @@ static rpmRC doReadRC(const char * urlfn)
 	    gotit = 0;
 
 	    for (i = 0; i < RPM_MACHTABLE_COUNT; i++) {
-		if (!strncmp(tables[i].key, s, strlen(tables[i].key)))
+		if (rstreqn(tables[i].key, s, strlen(tables[i].key)))
 		    break;
 	    }
 
@@ -573,20 +575,20 @@ static rpmRC doReadRC(const char * urlfn)
 		const char *rest = s + strlen(tables[i].key);
 		if (*rest == '_') rest++;
 
-		if (!strcmp(rest, "compat")) {
+		if (rstreq(rest, "compat")) {
 		    if (machCompatCacheAdd(se, fn, linenum,
 						&tables[i].cache))
 			goto exit;
 		    gotit = 1;
 		} else if (tables[i].hasTranslate &&
-			   !strcmp(rest, "translate")) {
+			   rstreq(rest, "translate")) {
 		    if (addDefault(&tables[i].defaults,
 				   &tables[i].defaultsLength,
 				   se, fn, linenum))
 			goto exit;
 		    gotit = 1;
 		} else if (tables[i].hasCanon &&
-			   !strcmp(rest, "canon")) {
+			   rstreq(rest, "canon")) {
 		    if (addCanon(&tables[i].canons, &tables[i].canonsLength,
 				 se, fn, linenum))
 			goto exit;
@@ -620,9 +622,9 @@ static rpmRC rpmPlatform(const char * platform)
     ssize_t blen = 0;
     int init_platform = 0;
     char * p, * pe;
-    int rc;
+    rpmRC rc;
 
-    rc = rpmioSlurp(platform, &b, &blen);
+    rc = (rpmioSlurp(platform, &b, &blen) == 0) ? RPMRC_OK : RPMRC_FAIL;
 
     if (rc || b == NULL || blen <= 0) {
 	rc = RPMRC_FAIL;
@@ -666,7 +668,7 @@ static rpmRC rpmPlatform(const char * platform)
 	while (*p && !(*p == '-' || isspace(*p)))
 	    p++;
 	if (*p != '-') {
-	    if (*p != '\0') *p++ = '\0';
+	    if (*p != '\0') *p = '\0';
 	    os = vendor;
 	    vendor = "unknown";
 	} else {
@@ -682,7 +684,7 @@ static rpmRC rpmPlatform(const char * platform)
 		while (*p && !(*p == '-' || isspace(*p)))
 		    p++;
 	    }
-	    if (*p != '\0') *p++ = '\0';
+	    if (*p != '\0') *p = '\0';
 	}
 
 	addMacro(NULL, "_host_cpu", NULL, cpu, -1);
@@ -808,8 +810,6 @@ static int is_athlon(void)
 	
 	cpuid (0, &eax, &ebx, &ecx, &edx);
 
- 	/* If you care about space, you can just check ebx, ecx and edx directly
- 	   instead of forming a string first and then doing a strcmp */
  	memset(vendor, 0, sizeof(vendor));
  	
  	for (i=0; i<4; i++)
@@ -819,13 +819,13 @@ static int is_athlon(void)
  	for (i=0; i<4; i++)
  		vendor[8+i] = (unsigned char) (ecx >>(8*i));
  		
- 	if (strncmp(vendor, "AuthenticAMD", 12) != 0)  
+ 	if (!rstreqn(vendor, "AuthenticAMD", 12))  
  		return 0;
 
 	return 1;
 }
 
-static int is_pentium3()
+static int is_pentium3(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
     char vendor[16];
@@ -834,7 +834,7 @@ static int is_pentium3()
     *((unsigned int *)&vendor[0]) = ebx;
     *((unsigned int *)&vendor[4]) = edx;
     *((unsigned int *)&vendor[8]) = ecx;
-    if (strncmp(vendor, "GenuineIntel", 12) != 0)
+    if (!rstreqn(vendor, "GenuineIntel", 12))
 	return 0;
     cpuid(1, &eax, &ebx, &ecx, &edx);
     family = (eax >> 8) & 0x0f;
@@ -845,18 +845,6 @@ static int is_pentium3()
 	    case 7:	// Pentium III, Pentium III Xeon (model 7)
 	    case 8:	// Pentium III, Pentium III Xeon, Celeron (model 8)
 	    case 9:	// Pentium M
-			/*
-			    Intel recently announced its new technology for mobile platforms,
-			    named Centrino, and presents it as a big advance in mobile PCs.
-			    One of the main part of Centrino consists in a brand new CPU,
-			    the Pentium M, codenamed Banias, that we'll study in this review.
-			    A particularity of this CPU is that it was designed for mobile platform
-			    exclusively, unlike previous mobile CPU (Pentium III-M, Pentium 4-M)
-			    that share the same micro-architecture as their desktop counterparts.
-			    The Pentium M introduces a new micro-architecture, adapted for mobility
-			    constraints, and that is halfway between the Pentium III and the Pentium 4.
-						    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-			*/
 	    case 10:	// Pentium III Xeon (model A)
 	    case 11:	// Pentium III (model B)
 		return 1;
@@ -864,7 +852,7 @@ static int is_pentium3()
     return 0;
 }
 
-static int is_pentium4()
+static int is_pentium4(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
     char vendor[16];
@@ -873,7 +861,7 @@ static int is_pentium4()
     *((unsigned int *)&vendor[0]) = ebx;
     *((unsigned int *)&vendor[4]) = edx;
     *((unsigned int *)&vendor[8]) = ecx;
-    if (strncmp(vendor, "GenuineIntel", 12) != 0)
+    if (!rstreqn(vendor, "GenuineIntel", 12))
 	return 0;
     cpuid(1, &eax, &ebx, &ecx, &edx);
     family = (eax >> 8) & 0x0f;
@@ -892,12 +880,10 @@ static int is_pentium4()
     return 0;
 }
 
-static int is_geode()
+static int is_geode(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
     char vendor[16];
-    /* If you care about space, you can just check ebx, ecx and edx directly
-       instead of forming a string first and then doing a strcmp */
     memset(vendor, 0, sizeof(vendor));
     
     cpuid(0, &eax, &ebx, &ecx, &edx);
@@ -905,7 +891,7 @@ static int is_geode()
     *((unsigned int *)&vendor[0]) = ebx;
     *((unsigned int *)&vendor[4]) = edx;
     *((unsigned int *)&vendor[8]) = ecx;
-    if (strncmp(vendor, "AuthenticAMD", 12) != 0)  
+    if (!rstreqn(vendor, "AuthenticAMD", 12))
         return 0;
     cpuid(1, &eax, &ebx, &ecx, &edx);
     family = (eax >> 8) & 0x0f;
@@ -920,12 +906,43 @@ static int is_geode()
 }
 #endif
 
-#if defined(__linux__) && defined(__powerpc__)
-static jmp_buf mfspr_jmpbuf;
 
-static void mfspr_ill(int notused)
+#if defined(__linux__)
+/**
+ * Populate rpmat structure with parsed info from /proc/self/auxv
+ */
+static void parse_auxv(void)
 {
-    longjmp(mfspr_jmpbuf, -1);
+    static int oneshot = 1;
+
+    if (oneshot) {
+	rpmat.platform = "";
+	int fd = open("/proc/self/auxv", O_RDONLY);
+
+	if (fd == -1) {
+	    rpmlog(RPMLOG_WARNING,
+		   _("Failed to read auxiliary vector, /proc not mounted?\n"));
+            return;
+	} else {
+	    ElfW(auxv_t) auxv;
+	    while (read(fd, &auxv, sizeof(auxv)) == sizeof(auxv)) {
+                switch (auxv.a_type)
+                {
+                    case AT_NULL:
+                        break;
+                    case AT_PLATFORM:
+                        rpmat.platform = strdup((char *) auxv.a_un.a_val);
+                        break;
+                    case AT_HWCAP:
+                        rpmat.hwcap = auxv.a_un.a_val;
+                        break;
+                }
+	    }
+	    close(fd);
+	}
+	oneshot = 0; /* only try once even if it fails */
+    }
+    return;
 }
 #endif
 
@@ -940,18 +957,22 @@ static void defaultMachine(const char ** arch,
     canonEntry canon;
     int rc;
 
+#if defined(__linux__)
+    /* Populate rpmat struct with hw info */
+    parse_auxv();
+#endif
+
     while (!gotDefaults) {
 	if (!rpmPlatform(platform)) {
-	    char * s;
-	    s = rpmExpand("%{_host_cpu}", NULL);
+	    char * s = rpmExpand("%{_host_cpu}", NULL);
 	    if (s) {
 		rstrlcpy(un.machine, s, sizeof(un.machine));
-		s = _free(s);
+		free(s);
 	    }
 	    s = rpmExpand("%{_host_os}", NULL);
 	    if (s) {
 		rstrlcpy(un.sysname, s, sizeof(un.sysname));
-		s = _free(s);
+		free(s);
 	    }
 	    gotDefaults = 1;
 	    break;
@@ -960,88 +981,34 @@ static void defaultMachine(const char ** arch,
 	if (rc < 0) return;
 
 #if !defined(__linux__)
-#ifdef SNI
-	/* USUALLY un.sysname on sinix does start with the word "SINIX"
-	 * let's be absolutely sure
-	 */
-	strncpy(un.sysname, "SINIX", sizeof(un.sysname));
-#endif
-	if (!strcmp(un.sysname, "AIX")) {
+	if (rstreq(un.sysname, "AIX")) {
 	    strcpy(un.machine, __power_pc() ? "ppc" : "rs6000");
 	    sprintf(un.sysname,"aix%s.%s", un.version, un.release);
 	}
-	else if(!strcmp(un.sysname, "Darwin")) { 
+	else if(rstreq(un.sysname, "Darwin")) { 
 #ifdef __ppc__
 	    strcpy(un.machine, "ppc");
 #else ifdef __i386__
 	    strcpy(un.machine, "i386");
 #endif 
 	}
-	else if (!strcmp(un.sysname, "SunOS")) {
-	    if (!strncmp(un.release,"4", 1)) /* SunOS 4.x */ {
-		int fd;
-		for (fd = 0;
-		    (un.release[fd] != 0 && (fd < sizeof(un.release)));
-		    fd++) {
-		      if (!risdigit(un.release[fd]) && (un.release[fd] != '.')) {
-			un.release[fd] = 0;
-			break;
-		      }
-		    }
-		    sprintf(un.sysname,"sunos%s",un.release);
-	    }
-
-	    else /* Solaris 2.x: n.x.x becomes n-3.x.x */
-		sprintf(un.sysname, "solaris%1d%s", atoi(un.release)-3,
-			un.release+1+(atoi(un.release)/10));
+	else if (rstreq(un.sysname, "SunOS")) {
+	    /* Solaris 2.x: n.x.x becomes n-3.x.x */
+	    sprintf(un.sysname, "solaris%1d%s", atoi(un.release)-3,
+		    un.release+1+(atoi(un.release)/10));
 
 	    /* Solaris on Intel hardware reports i86pc instead of i386
 	     * (at least on 2.6 and 2.8)
 	     */
-	    if (!strcmp(un.machine, "i86pc"))
+	    if (rstreq(un.machine, "i86pc"))
 		sprintf(un.machine, "i386");
 	}
-	else if (!strcmp(un.sysname, "HP-UX"))
+	else if (rstreq(un.sysname, "HP-UX"))
 	    /*make un.sysname look like hpux9.05 for example*/
 	    sprintf(un.sysname, "hpux%s", strpbrk(un.release, "123456789"));
-	else if (!strcmp(un.sysname, "OSF1"))
+	else if (rstreq(un.sysname, "OSF1"))
 	    /*make un.sysname look like osf3.2 for example*/
 	    sprintf(un.sysname, "osf%s", strpbrk(un.release, "123456789"));
-	else if (!strncmp(un.sysname, "IP", 2))
-	    un.sysname[2] = '\0';
-	else if (!strncmp(un.sysname, "SINIX", 5)) {
-	    sprintf(un.sysname, "sinix%s",un.release);
-	    if (!strncmp(un.machine, "RM", 2))
-		sprintf(un.machine, "mips");
-	}
-	else if ((!strncmp(un.machine, "34", 2) ||
-		!strncmp(un.machine, "33", 2)) && \
-		!strncmp(un.release, "4.0", 3))
-	{
-	    /* we are on ncr-sysv4 */
-	    char * prelid = NULL;
-	    FD_t fd = Fopen("/etc/.relid", "r.fdio");
-	    int gotit = 0;
-	    if (fd != NULL && !Ferror(fd)) {
-		chptr = xcalloc(1, 256);
-		{   int irelid = Fread(chptr, sizeof(*chptr), 256, fd);
-		    (void) Fclose(fd);
-		    /* example: "112393 RELEASE 020200 Version 01 OS" */
-		    if (irelid > 0) {
-			if ((prelid = strstr(chptr, "RELEASE "))){
-			    prelid += strlen("RELEASE ")+1;
-			    sprintf(un.sysname,"ncr-sysv4.%.*s",1,prelid);
-			    gotit = 1;
-			}
-		    }
-		}
-		chptr = _free (chptr);
-	    }
-	    if (!gotit)	/* parsing /etc/.relid file failed? */
-		strcpy(un.sysname,"ncr-sysv4");
-	    /* wrong, just for now, find out how to look for i586 later*/
-	    strcpy(un.machine,"i486");
-	}
 #endif	/* __linux__ */
 
 	/* get rid of the hyphens in the sysname */
@@ -1099,7 +1066,7 @@ static void defaultMachine(const char ** arch,
 #	endif	/* hpux */
 
 #	if defined(__linux__) && defined(__sparc__)
-	if (!strcmp(un.machine, "sparc")) {
+	if (rstreq(un.machine, "sparc")) {
 	    #define PERS_LINUX		0x00000000
 	    #define PERS_LINUX_32BIT	0x00800000
 	    #define PERS_LINUX32	0x00000008
@@ -1111,7 +1078,7 @@ static void defaultMachine(const char ** arch,
 	    if (oldpers != -1) {
 		if (personality(PERS_LINUX) != -1) {
 		    uname(&un);
-		    if (! strcmp(un.machine, "sparc64")) {
+		    if (rstreq(un.machine, "sparc64")) {
 			strcpy(un.machine, "sparcv9");
 			oldpers = PERS_LINUX32;
 		    }
@@ -1121,6 +1088,14 @@ static void defaultMachine(const char ** arch,
 	}
 #	endif	/* sparc*-linux */
 
+#	if defined(__linux__) && defined(__powerpc__)
+	{
+            int powerlvl;
+            if (sscanf(rpmat.platform, "power%d", &powerlvl) == 1 && powerlvl > 6)
+                strcpy(un.machine, "ppc64p7");
+        }
+#	endif	/* ppc64*-linux */
+
 #	if defined(__GNUC__) && defined(__alpha__)
 	{
 	    unsigned long amask, implver;
@@ -1149,9 +1124,9 @@ static void defaultMachine(const char ** arch,
 
 #	if defined(__linux__) && defined(__i386__)
 	{
-	    char class = (char) (RPMClass() | '0');
+	    char mclass = (char) (RPMClass() | '0');
 
-	    if ((class == '6' && is_athlon()) || class == '7')
+	    if ((mclass == '6' && is_athlon()) || mclass == '7')
 	    	strcpy(un.machine, "athlon");
 	    else if (is_pentium4())
 		strcpy(un.machine, "pentium4");
@@ -1159,8 +1134,8 @@ static void defaultMachine(const char ** arch,
 		strcpy(un.machine, "pentium3");
 	    else if (is_geode())
 		strcpy(un.machine, "geode");
-	    else if (strchr("3456", un.machine[1]) && un.machine[1] != class)
-		un.machine[1] = class;
+	    else if (strchr("3456", un.machine[1]) && un.machine[1] != mclass)
+		un.machine[1] = mclass;
 	}
 #	endif
 
@@ -1194,7 +1169,7 @@ const char * rpmGetVarArch(int var, const char * arch)
     if (arch) {
 	next = &values[var];
 	while (next) {
-	    if (next->arch && !strcmp(next->arch, arch)) return next->value;
+	    if (next->arch && rstreq(next->arch, arch)) return next->value;
 	    next = next->next;
 	}
     }
@@ -1217,7 +1192,7 @@ static void rpmSetVarArch(int var, const char * val, const char * arch)
     if (next->value) {
 	if (arch) {
 	    while (next->next) {
-		if (next->arch && !strcmp(next->arch, arch)) break;
+		if (next->arch && rstreq(next->arch, arch)) break;
 		next = next->next;
 	    }
 	} else {
@@ -1227,7 +1202,7 @@ static void rpmSetVarArch(int var, const char * val, const char * arch)
 	    }
 	}
 
-	if (next->arch && arch && !strcmp(next->arch, arch)) {
+	if (next->arch && arch && rstreq(next->arch, arch)) {
 	    next->value = _free(next->value);
 	    next->arch = _free(next->arch);
 	} else if (next->arch || arch) {
@@ -1244,7 +1219,7 @@ static void rpmSetVarArch(int var, const char * val, const char * arch)
     next->arch = (arch ? xstrdup(arch) : NULL);
 }
 
-void rpmSetTables(int archTable, int osTable)
+static void rpmSetTables(int archTable, int osTable)
 {
     const char * arch, * os;
 
@@ -1263,8 +1238,10 @@ void rpmSetTables(int archTable, int osTable)
 
 int rpmMachineScore(int type, const char * name)
 {
-    machEquivInfo info = machEquivSearch(&tables[type].equiv, name);
-    return (info != NULL ? info->score : 0);
+    machEquivInfo info = NULL;
+    if (name)
+	info = machEquivSearch(&tables[type].equiv, name);
+    return info ? info->score : 0;
 }
 
 int rpmIsKnownArch(const char *name)
@@ -1272,7 +1249,7 @@ int rpmIsKnownArch(const char *name)
     canonEntry canon = lookupInCanonTable(name,
 			tables[RPM_MACHTABLE_INSTARCH].canons,
 			tables[RPM_MACHTABLE_INSTARCH].canonsLength);
-    return (canon != NULL || strcmp(name, "noarch") == 0);
+    return (canon != NULL || rstreq(name, "noarch"));
 }
 
 /** \ingroup rpmrc
@@ -1310,13 +1287,13 @@ static void rpmSetMachine(const char * arch, const char * os)
     }
     if (os == NULL) return;	/* XXX can't happen */
 
-    if (!current[ARCH] || strcmp(arch, current[ARCH])) {
+    if (!current[ARCH] || !rstreq(arch, current[ARCH])) {
 	current[ARCH] = _free(current[ARCH]);
 	current[ARCH] = xstrdup(arch);
 	rebuildCompatTables(ARCH, host_cpu);
     }
 
-    if (!current[OS] || strcmp(os, current[OS])) {
+    if (!current[OS] || !rstreq(os, current[OS])) {
 	char * t = xstrdup(os);
 	current[OS] = _free(current[OS]);
 	/*
@@ -1327,7 +1304,7 @@ static void rpmSetMachine(const char * arch, const char * os)
 	 * XXX used by rpmInstallPackage->{os,arch}Okay->rpmMachineScore->
 	 * XXX to verify correct arch/os from headers.
 	 */
-	if (!strcmp(t, "linux"))
+	if (rstreq(t, "linux"))
 	    *t = 'L';
 	current[OS] = t;
 	
@@ -1424,7 +1401,7 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
     if (ca == NULL) {
 	const char *a = NULL;
 	defaultMachine(&a, NULL);
-	ca = (a) ? xstrdup(a) : NULL;
+	ca = xstrdup(a ? a : "(arch)");
     }
     for (x = 0; ca[x] != '\0'; x++)
 	ca[x] = rtolower(ca[x]);
@@ -1432,15 +1409,14 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
     if (co == NULL) {
 	const char *o = NULL;
 	defaultMachine(NULL, &o);
-	co = (o) ? xstrdup(o) : NULL;
+	co = xstrdup(o ? o : "(os)");
     }
     for (x = 0; co[x] != '\0'; x++)
 	co[x] = rtolower(co[x]);
 
     /* XXX For now, set canonical target to arch-os */
     if (ct == NULL) {
-	ct = xmalloc(strlen(ca) + sizeof("-") + strlen(co));
-	sprintf(ct, "%s-%s", ca, co);
+	rasprintf(&ct, "%s-%s", ca, co);
     }
 
 /*
@@ -1466,9 +1442,9 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
     if (canontarget)
 	*canontarget = ct;
     else
-	ct = _free(ct);
-    ca = _free(ca);
-    co = _free(co);
+	free(ct);
+    free(ca);
+    free(co);
 }
 
 void rpmFreeRpmrc(void)
@@ -1601,12 +1577,9 @@ exit:
 
 int rpmReadConfigFiles(const char * file, const char * target)
 {
-    mode_t mode = 0022;
-    /* Reset umask to its default umask(2) value. */
-    mode = umask(mode);
-
-    /* Force preloading of name service libraries in case we go chrooting */
+    /* Force preloading of dlopen()'ed libraries in case we go chrooting */
     (void) gethostbyname("localhost");
+    (void) rpmInitCrypto();
 
     /* Preset target macros */
    	/* FIX: target can be NULL */
@@ -1628,8 +1601,8 @@ int rpmReadConfigFiles(const char * file, const char * target)
     {	char *cpu = rpmExpand("%{_target_cpu}", NULL);
 	char *os = rpmExpand("%{_target_os}", NULL);
 	rpmSetMachine(cpu, os);
-	cpu = _free(cpu);
-	os = _free(os);
+	free(cpu);
+	free(os);
     }
 
 #ifdef WITH_LUA
@@ -1644,7 +1617,7 @@ int rpmShowRC(FILE * fp)
 {
     const struct rpmOption *opt;
     rpmds ds = NULL;
-    int i, xx;
+    int i;
     machEquivTable equivTable;
 
     /* the caller may set the build arch which should be printed here */
@@ -1692,7 +1665,7 @@ int rpmShowRC(FILE * fp)
     fprintf(fp, "\n");
 
     fprintf(fp, "Features supported by rpmlib:\n");
-    xx = rpmdsRpmlib(&ds, NULL);
+    rpmdsRpmlib(&ds, NULL);
     ds = rpmdsInit(ds);
     while (rpmdsNext(ds) >= 0) {
         const char * DNEVR = rpmdsDNEVR(ds);
