@@ -1,8 +1,4 @@
-/** \ingroup py_c
- * \file python/header-py.c
- */
-
-#include "system.h"
+#include "rpmsystem-py.h"
 
 #include <rpm/rpmlib.h>		/* rpmvercmp */
 #include <rpm/rpmtag.h>
@@ -11,9 +7,9 @@
 
 #include "header-py.h"
 #include "rpmds-py.h"
+#include "rpmfd-py.h"
 #include "rpmfi-py.h"
-
-#include "debug.h"
+#include "rpmtd-py.h"
 
 /** \ingroup python
  * \class Rpm
@@ -128,113 +124,97 @@
  * \name Class: rpm.hdr
  */
 
-/** \ingroup py_c
- */
 struct hdrObject_s {
     PyObject_HEAD
     Header h;
-} ;
+    HeaderIterator hi;
+};
 
-/** \ingroup py_c
- */
 static PyObject * hdrKeyList(hdrObject * s)
 {
-    PyObject * list, *o;
+    PyObject * keys;
     HeaderIterator hi;
-    rpmtd td = rpmtdNew();
+    rpmTagVal tag;
 
-    list = PyList_New(0);
+    keys = PyList_New(0);
+    if (!keys) {
+        return NULL;
+    }
 
     hi = headerInitIterator(s->h);
-    while (headerNext(hi, td)) {
-	rpmTag tag = rpmtdTag(td);
-        if (tag == HEADER_I18NTABLE) continue;
-
-	switch (rpmtdType(td)) {
-	case RPM_BIN_TYPE:
-	case RPM_INT32_TYPE:
-	case RPM_CHAR_TYPE:
-	case RPM_INT8_TYPE:
-	case RPM_INT16_TYPE:
-	case RPM_STRING_ARRAY_TYPE:
-	case RPM_STRING_TYPE:
-	    PyList_Append(list, o=PyInt_FromLong(tag));
-	    Py_DECREF(o);
-	    break;
-	case RPM_I18NSTRING_TYPE: /* hum.. ?`*/
-	case RPM_NULL_TYPE:
-	default:
-	    break;
-	}
-	rpmtdFreeData(td);
+    while ((tag = headerNextTag(hi)) != RPMTAG_NOT_FOUND) {
+	PyObject *to = PyInt_FromLong(tag);
+        if (!to) {
+            headerFreeIterator(hi);
+            Py_DECREF(keys);
+            return NULL;
+        }
+	PyList_Append(keys, to);
+	Py_DECREF(to);
     }
     headerFreeIterator(hi);
-    rpmtdFree(td);
 
-    return list;
+    return keys;
 }
 
-/** \ingroup py_c
- */
-static PyObject * hdrUnload(hdrObject * s, PyObject * args, PyObject *keywords)
+static PyObject * hdrAsBytes(hdrObject * s, int legacy)
 {
-    char * buf;
-    PyObject * rc;
-    int len, legacy = 0;
-    Header h;
-    static char *kwlist[] = { "legacyHeader", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, keywords, "|i", kwlist, &legacy))
-	return NULL;
-
-    h = headerLink(s->h);
+    PyObject *res = NULL;
+    char *buf = NULL;
+    unsigned int len;
+    Header h = headerLink(s->h);
+   
     /* XXX this legacy switch is a hack, needs to be removed. */
     if (legacy) {
 	h = headerCopy(s->h);	/* XXX strip region tags, etc */
 	headerFree(s->h);
     }
-    len = headerSizeof(h, 0);
-    buf = headerUnload(h);
+    buf = headerExport(h, &len);
     h = headerFree(h);
 
     if (buf == NULL || len == 0) {
 	PyErr_SetString(pyrpmError, "can't unload bad header\n");
-	return NULL;
+    } else {
+	res = PyBytes_FromStringAndSize(buf, len);
     }
+    free(buf);
+    return res;
+}
 
-    rc = PyString_FromStringAndSize(buf, len);
-    buf = _free(buf);
+static PyObject * hdrUnload(hdrObject * s, PyObject * args, PyObject *keywords)
+{
+    int legacy = 0;
+    static char *kwlist[] = { "legacyHeader", NULL};
 
-    return rc;
+    if (!PyArg_ParseTupleAndKeywords(args, keywords, "|i", kwlist, &legacy))
+	return NULL;
+
+    return hdrAsBytes(s, legacy);
 }
 
-/** \ingroup py_c
- */
 static PyObject * hdrExpandFilelist(hdrObject * s)
 {
+    DEPRECATED_METHOD("use hdr.convert() instead");
     headerConvert(s->h, HEADERCONV_EXPANDFILELIST);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
-/** \ingroup py_c
- */
 static PyObject * hdrCompressFilelist(hdrObject * s)
 {
+    DEPRECATED_METHOD("use hdr.convert() instead");
     headerConvert(s->h, HEADERCONV_COMPRESSFILELIST);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 /* make a header with _all_ the tags we need */
-/** \ingroup py_c
- */
-static void mungeFilelist(Header h)
+static PyObject * hdrFullFilelist(hdrObject * s)
 {
     rpmtd fileNames = rpmtdNew();
+    Header h = s->h;
 
+    DEPRECATED_METHOD("obsolete method");
     if (!headerIsEntry (h, RPMTAG_BASENAMES)
 	|| !headerIsEntry (h, RPMTAG_DIRNAMES)
 	|| !headerIsEntry (h, RPMTAG_DIRINDEXES))
@@ -246,23 +226,13 @@ static void mungeFilelist(Header h)
 	rpmtdFreeData(fileNames);
     }
     rpmtdFree(fileNames);
+
+    Py_RETURN_NONE;
 }
 
-/** \ingroup py_c
- */
-static PyObject * hdrFullFilelist(hdrObject * s)
+static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
 {
-    mungeFilelist (s->h);
-
-    Py_INCREF(Py_None);
-    return Py_None;
-}
-
-/** \ingroup py_c
- */
-static PyObject * hdrSprintf(hdrObject * s, PyObject * args, PyObject * kwds)
-{
-    char * fmt;
+    const char * fmt;
     char * r;
     errmsg_t err;
     PyObject * result;
@@ -278,16 +248,85 @@ static PyObject * hdrSprintf(hdrObject * s, PyObject * args, PyObject * kwds)
     }
 
     result = Py_BuildValue("s", r);
-    r = _free(r);
+    free(r);
 
     return result;
 }
 
-/**
- */
-static int hdr_compare(hdrObject * a, hdrObject * b)
+static PyObject *hdrIsSource(hdrObject *s)
 {
-    return rpmVersionCompare(a->h, b->h);
+    return PyBool_FromLong(headerIsSource(s->h));
+}
+
+static int hdrContains(hdrObject *s, PyObject *pytag)
+{
+    rpmTagVal tag;
+    if (!tagNumFromPyObject(pytag, &tag)) return -1;
+
+    return headerIsEntry(s->h, tag);
+}
+
+static PyObject *hdrConvert(hdrObject *self, PyObject *args, PyObject *kwds)
+{
+    char *kwlist[] = {"op", NULL};
+    int op = -1;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &op)) {
+        return NULL;
+    }
+    return PyBool_FromLong(headerConvert(self->h, op));
+}
+
+static PyObject * hdrWrite(hdrObject *s, PyObject *args, PyObject *kwds)
+{
+    char *kwlist[] = { "file", "magic", NULL };
+    int magic = HEADER_MAGIC_YES;
+    rpmfdObject *fdo = NULL;
+    int rc;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&|i", kwlist,
+				     rpmfdFromPyObject, &fdo, &magic))
+	return NULL;
+
+    Py_BEGIN_ALLOW_THREADS;
+    rc = headerWrite(rpmfdGetFd(fdo), s->h,
+		     magic ? HEADER_MAGIC_YES : HEADER_MAGIC_NO);
+    Py_END_ALLOW_THREADS;
+
+    if (rc) PyErr_SetFromErrno(PyExc_IOError);
+    Py_XDECREF(fdo); /* avoid messing up errno with file close  */
+    if (rc) return NULL;
+
+    Py_RETURN_NONE;
+}
+
+/*
+ * Just a backwards-compatibility dummy, the arguments are not looked at
+ * or used. TODO: push this over to python side...
+ */
+static PyObject * hdr_fiFromHeader(PyObject * s, PyObject * args, PyObject * kwds)
+{
+    return PyObject_CallFunctionObjArgs((PyObject *) &rpmfi_Type, s, NULL);
+}
+
+/* Backwards compatibility. Flags argument is just a dummy and discarded. */
+static PyObject * hdr_dsFromHeader(PyObject * s, PyObject * args, PyObject * kwds)
+{
+    rpmTagVal tag = RPMTAG_REQUIRENAME;
+    rpmsenseFlags flags = 0;
+    char * kwlist[] = {"to", "flags", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&i:dsFromHeader", kwlist,
+            tagNumFromPyObject, &tag, &flags))
+        return NULL;
+
+    return PyObject_CallFunction((PyObject *) &rpmds_Type,
+                                 "(Oi)", s, tag);
+}
+
+static PyObject * hdr_dsOfHeader(PyObject * s)
+{
+    return PyObject_CallFunction((PyObject *) &rpmds_Type,
+                                 "(Oi)", s, RPMTAG_NEVR);
 }
 
 static long hdr_hash(PyObject * h)
@@ -295,8 +334,17 @@ static long hdr_hash(PyObject * h)
     return (long) h;
 }
 
-/** \ingroup py_c
- */
+static PyObject * hdr_reduce(hdrObject *s)
+{
+    PyObject *res = NULL;
+    PyObject *blob = hdrAsBytes(s, 0);
+    if (blob) {
+	res = Py_BuildValue("O(O)", Py_TYPE(s), blob);
+	Py_DECREF(blob);
+    }
+    return res;
+}
+
 static struct PyMethodDef hdr_methods[] = {
     {"keys",		(PyCFunction) hdrKeyList,	METH_NOARGS,
 	NULL },
@@ -308,171 +356,302 @@ static struct PyMethodDef hdr_methods[] = {
 	NULL },
     {"fullFilelist",	(PyCFunction) hdrFullFilelist,	METH_NOARGS,
 	NULL },
-    {"sprintf",		(PyCFunction) hdrSprintf,	METH_VARARGS|METH_KEYWORDS,
+    {"convert",		(PyCFunction) hdrConvert,	METH_VARARGS|METH_KEYWORDS,
+	NULL },
+    {"format",		(PyCFunction) hdrFormat,	METH_VARARGS|METH_KEYWORDS,
+	NULL },
+    {"sprintf",		(PyCFunction) hdrFormat,	METH_VARARGS|METH_KEYWORDS,
+	NULL },
+    {"isSource",	(PyCFunction)hdrIsSource,	METH_NOARGS, 
+	NULL },
+    {"write",		(PyCFunction)hdrWrite,		METH_VARARGS|METH_KEYWORDS,
 	NULL },
-
     {"dsOfHeader",	(PyCFunction)hdr_dsOfHeader,	METH_NOARGS,
 	NULL},
     {"dsFromHeader",	(PyCFunction)hdr_dsFromHeader,	METH_VARARGS|METH_KEYWORDS,
 	NULL},
     {"fiFromHeader",	(PyCFunction)hdr_fiFromHeader,	METH_VARARGS|METH_KEYWORDS,
 	NULL},
+    {"__reduce__",	(PyCFunction)hdr_reduce,	METH_NOARGS,
+	NULL},
 
     {NULL,		NULL}		/* sentinel */
 };
 
-/** \ingroup py_c
- */
+/* TODO: permit keyring check + retrofits on copy/load */
+static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
+{
+    PyObject *obj = NULL;
+    rpmfdObject *fdo = NULL;
+    Header h = NULL;
+    char *kwlist[] = { "obj", NULL };
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O", kwlist, &obj)) {
+	return NULL;
+    }
+
+    if (obj == NULL) {
+	h = headerNew();
+    } else if (CAPSULE_CHECK(obj)) {
+	h = CAPSULE_EXTRACT(obj, "rpm._C_Header");
+    } else if (hdrObject_Check(obj)) {
+	h = headerCopy(((hdrObject*) obj)->h);
+    } else if (PyBytes_Check(obj)) {
+	h = headerCopyLoad(PyBytes_AsString(obj));
+    } else if (rpmfdFromPyObject(obj, &fdo)) {
+	Py_BEGIN_ALLOW_THREADS;
+	h = headerRead(rpmfdGetFd(fdo), HEADER_MAGIC_YES);
+	Py_END_ALLOW_THREADS;
+	Py_XDECREF(fdo);
+    } else {
+    	PyErr_SetString(PyExc_TypeError, "header, blob or file expected");
+	return NULL;
+    }
+
+    if (h == NULL) {
+	PyErr_SetString(pyrpmError, "bad header");
+	return NULL;
+    }
+    
+    return hdr_Wrap(subtype, h);
+}
+
 static void hdr_dealloc(hdrObject * s)
 {
     if (s->h) headerFree(s->h);
-    PyObject_Del(s);
+    Py_TYPE(s)->tp_free((PyObject *)s);
 }
 
-/** \ingroup py_c
- */
-rpmTag tagNumFromPyObject (PyObject *item)
+static PyObject * hdr_iternext(hdrObject *s)
 {
-    char * str;
+    PyObject *res = NULL;
+    rpmTagVal tag;
+
+    if (s->hi == NULL) s->hi = headerInitIterator(s->h);
+
+    if ((tag = headerNextTag(s->hi)) != RPMTAG_NOT_FOUND) {
+	res = PyInt_FromLong(tag);
+    } else {
+	s->hi = headerFreeIterator(s->hi);
+    }
+    return res;
+}
+
+int utf8FromPyObject(PyObject *item, PyObject **str)
+{
+    PyObject *res = NULL;
+    if (PyBytes_Check(item)) {
+	Py_XINCREF(item);
+	res = item;
+    } else if (PyUnicode_Check(item)) {
+	res = PyUnicode_AsUTF8String(item);
+    }
+    if (res == NULL) return 0;
+
+    *str = res;
+    return 1;
+}
+
+int tagNumFromPyObject (PyObject *item, rpmTagVal *tagp)
+{
+    rpmTagVal tag = RPMTAG_NOT_FOUND;
+    PyObject *str = NULL;
 
     if (PyInt_Check(item)) {
-	return PyInt_AsLong(item);
-    } else if (PyString_Check(item)) {
-	str = PyString_AsString(item);
-	return rpmTagGetValue(str);
+	/* XXX we should probably validate tag numbers too */
+	tag = PyInt_AsLong(item);
+    } else if (utf8FromPyObject(item, &str)) {
+	tag = rpmTagGetValue(PyBytes_AsString(str));
+	Py_DECREF(str);
+    } else {
+	PyErr_SetString(PyExc_TypeError, "expected a string or integer");
+	return 0;
     }
-    return RPMTAG_NOT_FOUND;
+    if (tag == RPMTAG_NOT_FOUND) {
+	PyErr_SetString(PyExc_ValueError, "unknown header tag");
+	return 0;
+    }
+
+    *tagp = tag;
+    return 1;
 }
 
-/** \ingroup py_c
- */
-static PyObject * hdr_subscript(hdrObject * s, PyObject * item)
+static PyObject * hdrGetTag(Header h, rpmTagVal tag)
 {
-    rpmTagType tagtype, type;
-    rpmTag tag = RPMTAG_NOT_FOUND;
-    rpm_count_t count, i;
-    rpm_data_t data;
-    PyObject * o, * metao;
-    char ** stringArray;
-    int forceArray = 0;
+    PyObject *res = NULL;
     struct rpmtd_s td;
 
-    tag = tagNumFromPyObject (item);
-    if (tag == RPMTAG_NOT_FOUND) {
-	PyErr_SetString(PyExc_KeyError, "unknown header tag");
-	return NULL;
-    }
-
-    tagtype = rpmTagGetType(tag); 
-    forceArray = (tagtype & RPM_MASK_RETURN_TYPE) == RPM_ARRAY_RETURN_TYPE;
-
-    /* Retrieve data from extension or header. */
-    if (!headerGet(s->h, tag, &td, HEADERGET_EXT)) {
-	if (forceArray) {
-	    o = PyList_New(0);
-	} else {
-	    o = Py_None;
-	    Py_INCREF(o);
-	}
-	return o;
-    }
-
-    count = td.count;
-    data = td.data;
-    type = td.type;
-
-    switch (type) {
-    case RPM_BIN_TYPE:
-	o = PyString_FromStringAndSize(data, count);
-	break;
-
-    case RPM_INT32_TYPE:
-	if (count != 1 || forceArray) {
-	    metao = PyList_New(0);
-	    for (i = 0; i < count; i++) {
-		o = PyInt_FromLong(((int *) data)[i]);
-		PyList_Append(metao, o);
-		Py_DECREF(o);
-	    }
-	    o = metao;
-	} else {
-	    o = PyInt_FromLong(*((int *) data));
-	}
-	break;
-
-    case RPM_CHAR_TYPE:
-    case RPM_INT8_TYPE:
-	if (count != 1 || forceArray) {
-	    metao = PyList_New(0);
-	    for (i = 0; i < count; i++) {
-		o = PyInt_FromLong(((char *) data)[i]);
-		PyList_Append(metao, o);
-		Py_DECREF(o);
-	    }
-	    o = metao;
-	} else {
-	    o = PyInt_FromLong(*((char *) data));
-	}
-	break;
-
-    case RPM_INT16_TYPE:
-	if (count != 1 || forceArray) {
-	    metao = PyList_New(0);
-	    for (i = 0; i < count; i++) {
-		o = PyInt_FromLong(((short *) data)[i]);
-		PyList_Append(metao, o);
-		Py_DECREF(o);
-	    }
-	    o = metao;
-	} else {
-	    o = PyInt_FromLong(*((short *) data));
-	}
-	break;
-
-    case RPM_STRING_ARRAY_TYPE:
-	stringArray = data;
-
-	metao = PyList_New(0);
-	for (i = 0; i < count; i++) {
-	    o = PyString_FromString(stringArray[i]);
-	    PyList_Append(metao, o);
-	    Py_DECREF(o);
-	}
-	o = metao;
-	break;
-
-    case RPM_STRING_TYPE:
-    case RPM_I18NSTRING_TYPE:
-	if (count != 1 || forceArray) {
-	    stringArray = data;
-
-	    metao = PyList_New(0);
-	    for (i=0; i < count; i++) {
-		o = PyString_FromString(stringArray[i]);
-		PyList_Append(metao, o);
-		Py_DECREF(o);
-	    }
-	    o = metao;
-	} else {
-	    o = PyString_FromString(data);
-	}
-	break;
-
-    default:
-	PyErr_SetString(PyExc_TypeError, "unsupported type in header");
-	return NULL;
+    (void) headerGet(h, tag, &td, HEADERGET_EXT);
+    if (rpmtdGetFlags(&td) & RPMTD_INVALID) {
+	PyErr_SetString(pyrpmError, "invalid header data");
+    } else {
+	/* rpmtd_AsPyobj() knows how to handle empty containers and all */
+	res = rpmtd_AsPyobj(&td);
     }
     rpmtdFreeData(&td);
-
-    return o;
+    return res;
 }
 
-static PyObject * hdr_getattro(PyObject * o, PyObject * n)
+static int validItem(rpmTagClass tclass, PyObject *item)
 {
-    PyObject * res;
-    res = PyObject_GenericGetAttr(o, n);
-    if (res == NULL)
-	res = hdr_subscript((hdrObject *)o, n);
+    int rc;
+
+    switch (tclass) {
+    case RPM_NUMERIC_CLASS:
+	rc = (PyLong_Check(item) || PyInt_Check(item));
+	break;
+    case RPM_STRING_CLASS:
+	rc = (PyBytes_Check(item) || PyUnicode_Check(item));
+	break;
+    case RPM_BINARY_CLASS:
+	rc = PyBytes_Check(item);
+	break;
+    default:
+	rc = 0;
+	break;
+    }
+    return rc;
+}
+
+static int validData(rpmTagVal tag, rpmTagType type, rpmTagReturnType retype, PyObject *value)
+{
+    rpmTagClass tclass = rpmTagGetClass(tag);
+    int valid = 1;
+    
+    if (retype == RPM_SCALAR_RETURN_TYPE) {
+	valid = validItem(tclass, value);
+    } else if (retype == RPM_ARRAY_RETURN_TYPE && PyList_Check(value)) {
+	/* python lists can contain arbitrary objects, validate each item */
+	Py_ssize_t len = PyList_Size(value);
+	if (len == 0)
+	    valid = 0;
+	for (Py_ssize_t i = 0; i < len; i++) {
+	    PyObject *item = PyList_GetItem(value, i);
+	    if (!validItem(tclass, item)) {
+		valid = 0;
+		break;
+	    }
+	}
+    } else {
+	valid = 0;
+    }
+    return valid;
+}
+
+static int hdrAppendItem(Header h, rpmTagVal tag, rpmTagType type, PyObject *item)
+{
+    int rc = 0;
+
+    switch (type) {
+    case RPM_I18NSTRING_TYPE: /* XXX this needs to be handled separately */
+    case RPM_STRING_TYPE:
+    case RPM_STRING_ARRAY_TYPE: {
+	PyObject *str = NULL;
+	if (utf8FromPyObject(item, &str)) 
+	    rc = headerPutString(h, tag, PyBytes_AsString(str));
+	Py_XDECREF(str);
+	} break;
+    case RPM_BIN_TYPE: {
+	uint8_t *val = (uint8_t *) PyBytes_AsString(item);
+	rpm_count_t len = PyBytes_Size(item);
+	rc = headerPutBin(h, tag, val, len);
+	} break;
+    case RPM_INT64_TYPE: {
+	uint64_t val = PyInt_AsUnsignedLongLongMask(item);
+	rc = headerPutUint64(h, tag, &val, 1);
+	} break;
+    case RPM_INT32_TYPE: {
+	uint32_t val = PyInt_AsUnsignedLongMask(item);
+	rc = headerPutUint32(h, tag, &val, 1);
+	} break;
+    case RPM_INT16_TYPE: {
+	uint16_t val = PyInt_AsUnsignedLongMask(item);
+	rc = headerPutUint16(h, tag, &val, 1);
+	} break;
+    case RPM_INT8_TYPE:
+    case RPM_CHAR_TYPE: {
+	uint8_t val = PyInt_AsUnsignedLongMask(item);
+	rc = headerPutUint8(h, tag, &val, 1);
+	} break;
+    default:
+	PyErr_SetString(PyExc_TypeError, "unhandled datatype");
+    }
+    return rc;
+}
+
+static int hdrPutTag(Header h, rpmTagVal tag, PyObject *value)
+{
+    rpmTagType type = rpmTagGetTagType(tag);
+    rpmTagReturnType retype = rpmTagGetReturnType(tag);
+    int rc = 0;
+
+    /* XXX this isn't really right (i18n strings etc) but for now ... */
+    if (headerIsEntry(h, tag)) {
+	PyErr_SetString(PyExc_TypeError, "tag already exists");
+	return rc;
+    }
+
+    /* validate all data before trying to insert */
+    if (!validData(tag, type, retype, value)) { 
+	PyErr_SetString(PyExc_TypeError, "invalid type for tag");
+	return 0;
+    }
+
+    if (retype == RPM_SCALAR_RETURN_TYPE) {
+	rc = hdrAppendItem(h, tag, type, value);
+    } else if (retype == RPM_ARRAY_RETURN_TYPE && PyList_Check(value)) {
+	Py_ssize_t len = PyList_Size(value);
+	for (Py_ssize_t i = 0; i < len; i++) {
+	    PyObject *item = PyList_GetItem(value, i);
+	    rc = hdrAppendItem(h, tag, type, item);
+	}
+    } else {
+	PyErr_SetString(PyExc_RuntimeError, "cant happen, right?");
+    }
+
+    return rc;
+}
+
+static PyObject * hdr_subscript(hdrObject * s, PyObject * item)
+{
+    rpmTagVal tag;
+
+    if (!tagNumFromPyObject(item, &tag)) return NULL;
+    return hdrGetTag(s->h, tag);
+}
+
+static int hdr_ass_subscript(hdrObject *s, PyObject *key, PyObject *value)
+{
+    rpmTagVal tag;
+    if (!tagNumFromPyObject(key, &tag)) return -1;
+
+    if (value == NULL) {
+	/* XXX should failure raise key error? */
+	headerDel(s->h, tag);
+    } else if (!hdrPutTag(s->h, tag, value)) {
+	return -1;
+    }
+    return 0;
+}
+
+static PyObject * hdr_getattro(hdrObject * s, PyObject * n)
+{
+    PyObject *res = PyObject_GenericGetAttr((PyObject *) s, n);
+    if (res == NULL) {
+	PyObject *type, *value, *traceback;
+	rpmTagVal tag;
+
+	/* Save and restore original exception if it's not a valid tag either */
+	PyErr_Fetch(&type, &value, &traceback);
+	if (tagNumFromPyObject(n, &tag)) {
+	    Py_XDECREF(type);
+	    Py_XDECREF(value);
+	    Py_XDECREF(traceback);
+	    res = hdrGetTag(s->h, tag);
+	} else {
+	    PyErr_Restore(type, value, traceback);
+	}
+    }
     return res;
 }
 
@@ -481,24 +660,30 @@ static int hdr_setattro(PyObject * o, PyObject * n, PyObject * v)
     return PyObject_GenericSetAttr(o, n, v);
 }
 
-/** \ingroup py_c
- */
 static PyMappingMethods hdr_as_mapping = {
 	(lenfunc) 0,			/* mp_length */
 	(binaryfunc) hdr_subscript,	/* mp_subscript */
-	(objobjargproc)0,		/* mp_ass_subscript */
+	(objobjargproc)hdr_ass_subscript,/* mp_ass_subscript */
+};
+
+static PySequenceMethods hdr_as_sequence = {
+    0,				/* sq_length */
+    0,				/* sq_concat */
+    0,				/* sq_repeat */
+    0,				/* sq_item */
+    0,				/* sq_slice */
+    0,				/* sq_ass_item */
+    0,				/* sq_ass_slice */
+    (objobjproc)hdrContains,	/* sq_contains */
+    0,				/* sq_inplace_concat */
+    0,				/* sq_inplace_repeat */
 };
 
-/**
- */
 static char hdr_doc[] =
 "";
 
-/** \ingroup py_c
- */
 PyTypeObject hdr_Type = {
-	PyObject_HEAD_INIT(&PyType_Type)
-	0,				/* ob_size */
+	PyVarObject_HEAD_INIT(&PyType_Type, 0)
 	"rpm.hdr",			/* tp_name */
 	sizeof(hdrObject),		/* tp_size */
 	0,				/* tp_itemsize */
@@ -506,10 +691,10 @@ PyTypeObject hdr_Type = {
 	0,				/* tp_print */
 	(getattrfunc) 0, 		/* tp_getattr */
 	0,				/* tp_setattr */
-	(cmpfunc) hdr_compare,		/* tp_compare */
+	0,				/* tp_compare */
 	0,				/* tp_repr */
 	0,				/* tp_as_number */
-	0,	 			/* tp_as_sequence */
+	&hdr_as_sequence,		/* tp_as_sequence */
 	&hdr_as_mapping,		/* tp_as_mapping */
 	hdr_hash,			/* tp_hash */
 	0,				/* tp_call */
@@ -517,15 +702,14 @@ PyTypeObject hdr_Type = {
 	(getattrofunc) hdr_getattro,	/* tp_getattro */
 	(setattrofunc) hdr_setattro,	/* tp_setattro */
 	0,				/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT,		/* tp_flags */
+	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,	/* tp_flags */
 	hdr_doc,			/* tp_doc */
-#if Py_TPFLAGS_HAVE_ITER
 	0,				/* tp_traverse */
 	0,				/* tp_clear */
 	0,				/* tp_richcompare */
 	0,				/* tp_weaklistoffset */
-	0,				/* tp_iter */
-	0,				/* tp_iternext */
+	PyObject_SelfIter,		/* tp_iter */
+	(iternextfunc)hdr_iternext,	/* tp_iternext */
 	hdr_methods,			/* tp_methods */
 	0,				/* tp_members */
 	0,				/* tp_getset */
@@ -536,147 +720,40 @@ PyTypeObject hdr_Type = {
 	0,				/* tp_dictoffset */
 	0,				/* tp_init */
 	0,				/* tp_alloc */
-	0,				/* tp_new */
+	hdr_new,			/* tp_new */
 	0,				/* tp_free */
 	0,				/* tp_is_gc */
-#endif
 };
 
-hdrObject * hdr_Wrap(Header h)
+PyObject * hdr_Wrap(PyTypeObject *subtype, Header h)
 {
-    hdrObject * hdr = PyObject_New(hdrObject, &hdr_Type);
+    hdrObject * hdr = (hdrObject *)subtype->tp_alloc(subtype, 0);
+    if (hdr == NULL) return NULL;
+
     hdr->h = headerLink(h);
-    return hdr;
-}
-
-Header hdrGetHeader(hdrObject * s)
-{
-    return s->h;
-}
-
-/**
- */
-PyObject * hdrLoad(PyObject * self, PyObject * args, PyObject * kwds)
-{
-    hdrObject * hdr;
-    char * obj;
-    Header h;
-    int len;
-    char * kwlist[] = {"headers", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s#", kwlist, &obj, &len))
-	return NULL;
-
-    /* copy is needed to avoid surprises from data swab in headerLoad(). */
-    h = headerCopyLoad(obj);
-    if (!h) {
-	if (errno == ENOMEM) {
-	    PyErr_SetString(pyrpmError, "out of memory");
-	} else {
-	    PyErr_SetString(pyrpmError, "bad header");
-	}
-	return NULL;
-    }
-    headerConvert(h, HEADERCONV_RETROFIT_V3);
-
-    hdr = hdr_Wrap(h);
-    h = headerFree(h);	/* XXX ref held by hdr */
-
     return (PyObject *) hdr;
 }
 
-/**
- */
-PyObject * rpmReadHeaders (FD_t fd)
+int hdrFromPyObject(PyObject *item, Header *hptr)
 {
-    PyObject * list;
-    Header h;
-    hdrObject * hdr;
-
-    if (!fd) {
-	PyErr_SetFromErrno(pyrpmError);
-	return NULL;
+    if (hdrObject_Check(item)) {
+	*hptr = ((hdrObject *) item)->h;
+	return 1;
+    } else {
+	PyErr_SetString(PyExc_TypeError, "header object expected");
+	return 0;
     }
-
-    list = PyList_New(0);
-    Py_BEGIN_ALLOW_THREADS
-    h = headerRead(fd, HEADER_MAGIC_YES);
-    Py_END_ALLOW_THREADS
-
-    while (h) {
-	headerConvert(h, HEADERCONV_RETROFIT_V3);
-	hdr = hdr_Wrap(h);
-	if (PyList_Append(list, (PyObject *) hdr)) {
-	    Py_DECREF(list);
-	    Py_DECREF(hdr);
-	    return NULL;
-	}
-	Py_DECREF(hdr);
-
-	h = headerFree(h);	/* XXX ref held by hdr */
-
-	Py_BEGIN_ALLOW_THREADS
-	h = headerRead(fd, HEADER_MAGIC_YES);
-	Py_END_ALLOW_THREADS
-    }
-
-    return list;
-}
-
-/**
- */
-PyObject * rpmHeaderFromFD(PyObject * self, PyObject * args, PyObject * kwds)
-{
-    FD_t fd;
-    int fileno;
-    PyObject * list;
-    char * kwlist[] = {"fd", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &fileno))
-	return NULL;
-
-    fd = fdDup(fileno);
-
-    list = rpmReadHeaders (fd);
-    Fclose(fd);
-
-    return list;
-}
-
-/**
- */
-PyObject * rpmHeaderFromFile(PyObject * self, PyObject * args, PyObject *kwds)
-{
-    char * filespec;
-    FD_t fd;
-    PyObject * list;
-    char * kwlist[] = {"file", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s", kwlist, &filespec))
-	return NULL;
-
-    fd = Fopen(filespec, "r.fdio");
-
-    if (!fd) {
-	PyErr_SetFromErrno(pyrpmError);
-	return NULL;
-    }
-
-    list = rpmReadHeaders (fd);
-    Fclose(fd);
-
-    return list;
 }
 
 /**
  * This assumes the order of list matches the order of the new headers, and
  * throws an exception if that isn't true.
  */
-int rpmMergeHeaders(PyObject * list, FD_t fd, int matchTag)
+static int rpmMergeHeaders(PyObject * list, FD_t fd, int matchTag)
 {
     Header h;
     HeaderIterator hi;
-    rpmTag newMatch, oldMatch;
+    rpmTagVal newMatch, oldMatch;
     hdrObject * hdr;
     rpm_count_t count = 0;
     int rc = 1; /* assume failure */
@@ -758,58 +835,9 @@ rpmMergeHeadersFromFD(PyObject * self, PyObject * args, PyObject * kwds)
 	return NULL;
     }
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
-/**
- */
-PyObject *
-rpmSingleHeaderFromFD(PyObject * self, PyObject * args, PyObject * kwds)
-{
-    FD_t fd;
-    int fileno;
-    off_t offset;
-    PyObject * tuple;
-    Header h;
-    char * kwlist[] = {"fd", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &fileno))
-	return NULL;
-
-    offset = lseek(fileno, 0, SEEK_CUR);
-
-    fd = fdDup(fileno);
-
-    if (!fd) {
-	PyErr_SetFromErrno(pyrpmError);
-	return NULL;
-    }
-
-    Py_BEGIN_ALLOW_THREADS
-    h = headerRead(fd, HEADER_MAGIC_YES);
-    Py_END_ALLOW_THREADS
-
-    Fclose(fd);
-
-    tuple = PyTuple_New(2);
-
-    if (h && tuple) {
-	PyTuple_SET_ITEM(tuple, 0, (PyObject *) hdr_Wrap(h));
-	PyTuple_SET_ITEM(tuple, 1, PyLong_FromLong(offset));
-	h = headerFree(h);
-    } else {
-	Py_INCREF(Py_None);
-	Py_INCREF(Py_None);
-	PyTuple_SET_ITEM(tuple, 0, Py_None);
-	PyTuple_SET_ITEM(tuple, 1, Py_None);
-    }
-
-    return tuple;
-}
-
-/**
- */
 PyObject * versionCompare (PyObject * self, PyObject * args, PyObject * kwds)
 {
     hdrObject * h1, * h2;
@@ -819,11 +847,9 @@ PyObject * versionCompare (PyObject * self, PyObject * args, PyObject * kwds)
 	    &h1, &hdr_Type, &h2))
 	return NULL;
 
-    return Py_BuildValue("i", hdr_compare(h1, h2));
+    return Py_BuildValue("i", rpmVersionCompare(h1->h, h2->h));
 }
 
-/**
- */
 static int compare_values(const char *str1, const char *str2)
 {
     if (!str1 && !str2)
@@ -837,7 +863,7 @@ static int compare_values(const char *str1, const char *str2)
 
 PyObject * labelCompare (PyObject * self, PyObject * args)
 {
-    char *v1, *r1, *v2, *r2;
+    const char *v1, *r1, *v2, *r2;
     const char *e1, *e2;
     int rc;
 
